/*
Murasaki - multiple genome global alignment program
Copyright (C) 2006-2007 Keio University
(Kris Popendorf) <comp@bio.keio.ac.jp> (2006)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

////////
// sequence defs
// Kris Popendorf
///////

#ifndef SEQUENCE__H
#define SEQUENCE__H

#include "cmultiset.h"

#include <iostream>
#include <vector>
#include <pstream.h>
#include <pstream.h>
#include <set>
#include <map>
#include <list>
#include <limits.h>
#include <functional>
#include <assert.h>
#include "itree.h"

#ifdef MURASAKI_SQLITE3
#include <sqlite3.h>
#endif

#ifdef MURASAKI_MYSQL
#include <mysql.h>
#endif

#ifdef MURASAKI_MPI
#include "mpi.h"
#define MURASAKI_MPI_REQ_SIZE 0
#define MURASAKI_MPI_REQ_LKUP_NOMATCH 1
#define MURASAKI_MPI_REQ_LKUP_MATCH 2
#define MURASAKI_MPI_REQ_QUIT  999
#endif

using namespace std;
using namespace Itree;

//in my world:
// C = 00
// A = 01
// T = 10
// G = 11
// isnt that cute? ~G==C and so on

#ifdef __linux__
#define WORDSIZE __WORDSIZE
#endif

#if defined (__FreeBSD__) || (defined (__APPLE__) && defined (__ppc__))
#define WORDSIZE 32
#endif

#if defined (__APPLE__) && defined (__ppc64__)
#define WORDSIZE 64
#endif

#define MODWORDSIZE(a)  ((a) & (WORDSIZE-1))

typedef unsigned long word; 
extern int _wordsize,hash_max; 
extern int max_hash_bits;//hash_max limits memory usage
extern int hash_size; //actually used size, might be less than max
extern int mismatch_penalty,match_score,max_credits;
extern bool gappingEnabled;
typedef word HashKey;

#ifdef MURASAKI_MPI
extern int mpi_id, mpi_procs;
#endif

class Sequence;
class BitSequence;
void set_seqtable(vector<Sequence*> seqs);

void initHashParams(int bits,BitSequence *pat);
void initConstants();

inline int highestBit(word t);
inline word lowN(int i);
inline word highN(int i);
inline word first2(word w){return w>>(WORDSIZE-2);}

int popCount(word w);
string wordToString(word w,int bits=WORDSIZE);
string wordToPattern(word w,int bits=WORDSIZE);

string hashCacheName(string name, BitSequence *pat);

//bits for sequence index table
typedef unsigned char SeqIdx;
extern map<Sequence*,SeqIdx> seq2idx;

//forward declarations...

extern int next_sequence_;
extern int base_patt_len;
extern word activeHash;

//inline function defs actually have to be in here if they're used in other .cc's...
const inline SeqIdx lookup_seqtable(Sequence* s) {
  assert(seq2idx.find(s)!=seq2idx.end()); //must be in there already
  return seq2idx[s];
}

inline int bitSeqCoords(int seqCoords){
  return abs(seqCoords)-1;
}

inline word lowN(int i){
  return ~(((word)-1)<<i); //lowest n bits are 1
}

inline word highN(int i){
  return ~((~(word)0)>>i);
}

void mpi_client_mode();

class HashMethod;
class Window;
class Anchor;

class BitSequence { //for packing dna strings into 2 bit chunks
 public:
  //accessor
  inline const vector<word>& words(){return _words;};
  inline int length(){return _length;}
  int wordsize();
  string asString() const;
  string asPattern() const;
  BitSequence* substr(int start,int stop);
  inline int readCode(int pos) const;
  inline word readWord(int pos) const;
  inline int readBitCount(){return bit_count;}
  inline int hashLength(){return _hashLength;}
  int compHashLength();
  pair<int,int>* localRegion(int pos); //find region local to pos
  int spaceRight(int pos); //matchable space right of pos
  int spaceLeft(int pos); //matchable space left of pos

  int cmpRight(int pos,BitSequence& target,int targetPos);
  int cmpLeft(int pos,BitSequence& target,int targetPos);

  //manipulate others
  void maskString(string &str) const;

  //operator
  bool equal(int pos,BitSequence& target,int targetPos,BitSequence& patt);
  BitSequence* extract(int pos,BitSequence &p);  

  //mutator
  void invert(); //flips to comp sequence
  
  //init
  void init(const string &str);
  void init(redi::ipstream &inf);
  void init(const BitSequence& a);

  //constructamalators
  BitSequence(redi::ipstream &inf); //from input stream
  BitSequence(const string &str); //from either GATC or 0101 (as mask: ie double each bit)
  BitSequence(const BitSequence& a); //copy
  BitSequence();//uninited

  friend bool operator==(const BitSequence& a,const BitSequence& b);
  friend class Window;
  friend class HashMethod;
  friend void initHashParams(int bits,BitSequence *pat);
  friend ostream& operator<<(ostream &os,const BitSequence &a);

  //oo public data. dont break it (yeah, as if the rest of this program as so indestructible)
  vector<pair<int,int> > matchRegions;

 protected:
  void compileHashFunc();
  int entropyCheck(vector<int> &entropy);
  void addWord(word w,int bits);
  vector<word> _words;
  int _length;
  int word_count,bit_count;
  int _hashLength;
  bool isPattern;

  //for fast hash computation:
  HashMethod* hasher;
};

class HashMethod {
 public:
  typedef vector<pair<int,int> > InputList;
  InputList inputs;
  HashMethod(HashMethod &a);
  HashMethod(BitSequence *s);

  static map<int,int> popmap;

  BitSequence *seq;
  int score,id;

  //scoring state
  vector<int> entropy,used;
  int sources,empties;
  word active;
  friend ostream& operator<<(ostream &os,const HashMethod &a);
  int align(int w); //propose an offset for a new input
  void finalize();
  int maxSources();
  void mutate();
  word hash(Window &w);
  int entropyCheck();
  void removeDuplicates();
  void pruneUseless();
 private:
  static int nextId;
};

class Location;

//tool for skimming bitsequences through a pattern. binds to BitSequence*
class Window { 
 public:
  BitSequence *seq,*pat;
  int pos,frame;
  void slide();
  void slide(int dist);
  bool equals(const Window &a) const;
  word eqbases(const Window &a) const;
  word hash();
  word mask;
  Window(BitSequence *s,int pos,BitSequence *p);
  Window(Location &v,BitSequence *p);
  Window(Anchor *a,BitSequence *p);
  friend ostream& operator<<(ostream &os, const Window &a);
  friend bool operator<(const Window &a,const Window &b);
  string prettyString();
  string toRawString();

 protected:
  void initslide(int dist);
  int frames;
  vector<word> buffer;
};

class Sequence {
 public:
  Sequence(); //empty sequence (ie: useless)
  Sequence(string); //from a file
  Sequence(const Sequence &obj); // duplicate

  string filename,name;
  int length();
  inline int getId(){return seqID;}
  word mclass();
  inline int extCoords(int mcoords)
    {return mcoords<0 ? -1-(rev->length()+mcoords):mcoords;}

  friend bool operator ==(const Sequence&,const Sequence&);
  BitSequence *fwd,*rev;

 protected:
  int seqID;
};

class Location {
 public:
  SeqIdx seqno;
  int pos;

  Sequence* seq() const;
  inline int seqId() const {return seq()->getId();}
  BitSequence* bitSeq();
  bool equal(const Location &a,BitSequence &patt) const;
  friend bool operator==(const Location &a,const Location &b);
  friend bool operator<(const Location &a,const Location &b);
  friend ostream& operator<<(ostream &os,const Location &a);
  
  Location(Sequence*,int);
  Location(SeqIdx, int);
  Location(); //this is dangerous. be careful with it.
};

typedef Location HashVal;
typedef list<Location> LocSubList;
typedef vector<LocSubList> LocList;
bool operator==(const HashVal &a,const HashVal &b);

class Hash {
 public:
  //abstract parts
  virtual void clear() = 0;
  virtual void add(const HashKey key,const HashVal &val) = 0;
  virtual void lookup(HashKey key,const HashVal &a,LocList &locList) = 0;
  virtual void lookup(HashKey key,LocList &locList) = 0;
  virtual bool emptyAt(const HashKey key) = 0;
  virtual unsigned sizeAt(const HashKey key) = 0;
 
  //for computing memory costs
  virtual const int new_bucket_cost() const = 0;
  virtual const int linear_cost() const = 0;
  virtual const int bucket_prep_cost() const = 0;

  // dump / load may be overloaded
  virtual void dump(ostream &os);
  virtual void load(istream &is);

  //non virtual stuff that will never change
  void writeHistogram(ostream &of);
  void writeDetailedHistogram(ostream &of);
  BitSequence* hashpat;
  bool sanityCheck();
  bool sanityCheck(word base);
  unsigned sizeAt(const HashKey key,const HashVal &a); //slow but compatible
  
  //the boring junk
  Hash(BitSequence *pat);
  virtual ~Hash() = 0; 
 protected:
  //nada
};

class ArrayHash : public Hash {
 public:
  //my own lovely bunch of coconuts
  typedef cmultiset<HashVal>*  MHash; //multihash
  typedef cmultiset<HashVal>::iterator HashIte;

  void clear();
  void add(const HashKey key,const HashVal &val);
  void lookup(HashKey key,const HashVal &a,LocList &locList);
  void lookup(HashKey key,LocList &locList); //this is just for debugging...
  bool emptyAt(const HashKey key);
  unsigned sizeAt(const HashKey key);

  ArrayHash(BitSequence *pat);

  void dump(ostream &os);
  void load(istream &is);

  const int new_bucket_cost() const {
    return sizeof(cmultiset<HashVal>*);
  }
  const int linear_cost() const {
    return sizeof(Location);
  }
  const int bucket_prep_cost() const {
    return sizeof(cmultiset<HashVal>);
  }

  bool rawSanityCheck();
  bool rawSanityCheck(word base);

 protected:
  void lookup(HashKey, const HashVal &, LocList &, bool);
  void pickup(LocList &locList, pair<HashIte,HashIte>);
  MHash *fasta;
  pair<HashIte,HashIte> activeRange;
};


class AnchorSet;
class Anchor;
typedef Anchor* Space;
typedef itree<int,Space> usedItree;
typedef usedItree::iterator Ancite;
typedef Interval<int> usedInt;

class Anchor {
 public:
  int start,stop; //in global coords (ie: -n to -1 and 1 to n)
  AnchorSet *parent;
  Sequence *seq;
  Ancite mapite;
  
  int length() const;
  int setId() const;
  Anchor(int _start,int _stop,AnchorSet *_parent,Sequence *_seq);
  int cmp(const Anchor &a) const;
  void coalesce(const Anchor &a);
  bool contains(const Anchor &a);
  bool sanityCheck() const;
  inline usedInt asInterval() const {return usedInt(start,stop);}
  
  friend bool operator<(const Anchor &a,const Anchor &b);

  string asString() const;
  friend ostream& operator<<(ostream &os, const Anchor &a);
};

struct ltSpace {
  inline bool operator()(const Space &a, const Space &b) const {
    return *a < *b;
  }
};

struct ltRegion : binary_function<pair<int,int>,pair<int,int>,bool> {
  inline bool operator()(const pair<int,int> &a, const pair<int,int> &b) const {
    return a.second<b.second;
  }
};

class AnchorSet {
 public:
  static int _nextId;
  int setId;
  word mclass;
  int hitCount;
  typedef map<word,int> HashCount;
  HashCount members;

  int length();
  void extend();
  vector<Space> spaces;
  void add(Space);
  void add(HashVal&,int length);
  AnchorSet();
  ~AnchorSet();
  void coalesce(const AnchorSet& neighbor);
  bool contains(const AnchorSet &a);
  bool overlaps(const AnchorSet &a);
  bool sanityCheck() const;
  double uniqueness() const;
  pair<int,int> bitscore() const;

  static AnchorSet* lookupSet(int id);

  string asString() const;
  friend ostream& operator<<(ostream &os, const AnchorSet &a);
 protected:
  static map<int,AnchorSet*> setMap;
};

class UsedMap {
 public:
  bool alreadyExists(const AnchorSet &a);
  bool merge(AnchorSet& a);
  void add(AnchorSet &a);
  void remove(AnchorSet *a); //deletes a!
  UsedMap(int seqs);
  UsedMap(const string filename); //load from file
  bool sanityCheck();
  string asString();
  int count() const;

  ostream& saveDetails(ostream &os,ostream &bitos); //make second cerr to skip
  void writeOut(ostream &os);
 protected:
  vector<usedItree > used;
  bool fetchSetIds(int seq,const Space &a,set<int> &out);
};

#ifdef MURASAKI_MPI
struct mpi_req{
  int type;
  HashKey key;
  HashVal val;
};
#endif

#endif

